import csv
from html.parser import HTMLParser
import logging

import numpy as np

from crossnetwork_processing_functions import create_GHOST_stage_logfile
from GHOST_standards import standard_parameters


class ParseAERONETFile(HTMLParser):

    '''class that handles parsing of an AERONET html file'''

    def __init__(self):
        super(ParseAERONETFile, self).__init__(convert_charrefs=True)

        self.have_title = False
        self.have_contact = False
        self.have_column_names = False

    def handle_starttag(self, tag, attrs):
        if tag == 'title':
            self.have_title = True

    def handle_endtag(self, tag):
        pass

    def handle_data(self, data):
        if (self.have_title) & (self.have_contact) & (self.have_column_names):
            data_split = data.strip("\r\n").split(',')
            self.current_n_columns = len(data_split)
            #if number of current line columns is < than number of data columns, do not process line
            if self.current_n_columns >= self.n_columns:
                #if number of current line columns is != to previous number of accepted columns, do not process line
                if (self.previous_n_columns == 0) or (self.previous_n_columns == self.current_n_columns):
                    for column_ii, column_name in enumerate(self.column_names):
                        self.var_data[column_name].append(data_split[column_ii])
                    self.previous_n_columns = self.current_n_columns

        elif (self.have_title) & (self.have_contact):
            self.column_names = data.strip("\r\n").replace(" ", "").split(',')
            self.n_columns = len(self.column_names)
            self.previous_n_columns = 0
            self.var_data = {k:[] for k in self.column_names}
            self.have_column_names = True

        elif self.have_title:
            if 'Contact:' in data:
                contact_str = data.strip("\r\n").replace(" ", "")
                #get contact name
                if "PI=" in contact_str:
                    self.contact_name = contact_str.split("PI=")[1].split(";")[0].replace("_", " ")
                else:
                    self.contact_name = np.NaN
                #get contact email address
                if "PIEmail=" in contact_str:
                    self.contact_email_address = contact_str.split("PIEmail=")[1].replace("_", " ")
                else:
                    self.contact_email_address = np.NaN
                #contact institution
                self.contact_institution = 'NASA'
                self.have_contact = True

                #handle case where contact line is not line broken and the column names directly follow
                if "AERONET_Site," in data:
                    column_name_line = data.strip("\r\n").replace(" ", "")
                    column_name_line = 'AERONET_Site,Date'+column_name_line.split('AERONET_Site,Date')[-1]
                    self.column_names = column_name_line.split(',')
                    self.n_columns = len(self.column_names)
                    self.previous_n_columns = 0
                    self.var_data = {k:[] for k in self.column_names}
                    self.have_column_names = True


# class ParseCAPMoNFile:
# 
#     '''class that handles parsing of a CAPMoN annual data file'''
#     
#     def __init__(self, filename):
# 
#         #write filename to self
#         self.filename = filename
# 
#         #get data year
#         self.data_year = self.filename.split('.csv')[-1][-4:]
# 
#         #define dictionaries to store metadata/data
#         self.metadata = {}
#         self.data = {}
# 
#     def read_file(self):
# 
#         #define variables to know what step of file read is currently active
#         file_header_section = False
#         qa_section = False
#         station_metadata_section = False
#         data_section = False
# 
#         #open file
#         with open(self.filename, 'rb', encoding='latin1') as f:
#             reader = csv.reader(f,delimiter=',',quotechar='"')
#                 
#             #read line by line
#             for row in reader:
# 
#                 #---------------------------------------------------------#
#                 #test if data section has changed between lines
#                 #---------------------------------------------------------#
# 
#                 #line starting file header section?
#                 if row[0] == "*FILE NAME // NOM DU FICHIER":
#                     file_header_section = True
#                     qa_section = False
#                     station_metadata_section = False
#                     data_section = False
# 
#                 #line starting qa section
#                 elif (row[0] == '*TABLE NAME // NOM DE LA TABLE') & (row[1] == "Data validity flags // Indicateurs de validité des données"):
#                     file_header_section = False
#                     qa_section = True
#                     station_metadata_section = False
#                     data_section = False
#                 
#                 #line starting station metadata section
#                 elif (row[0] == '*TABLE NAME // NOM DE LA TABLE') & (row[1] == "Site information"):
#                     file_header_section = False
#                     qa_section = False
#                     station_metadata_section = True
#                     data_section = False
# 
#                 #line starting data section
#                 elif (row[0] == '*TABLE NAME // NOM DE LA TABLE') & (row[1] == "CAPMoN_Ozone"):
#                     file_header_section = False
#                     qa_section = False
#                     station_metadata_section = False
#                     data_section = True
#                 
# 
#                 #---------------------------------------------------------#
#                 #parse line
#                 #---------------------------------------------------------#
# 
#                 #in file header section?
#                 if file_header_section:
#                     
#                     #contact institution
#                     if row[0] == "*ORGANIZATION NAME // NOM DE L'ORGANISATION":
#                         self.metadata['contact_institution'] = row[1].split("//")[0].strip(' "')
# 
#                     #contact email
#                     if row[0] == "*ORGANIZATION EMAIL // COURRIEL DE L'ORGANISATION":
#                         self.metadata['contact_email_address'] = row[1].split("//")[0].strip(' "')
# 
#                     #principal investigator name
#                     if row[0] == "*PRINCIPAL INVESTIGATOR NAME (LAST,FIRST) // NOM DU CHERCHEUR PRINCIPAL (NOM, PRÉNOM)":
#                         self.metadata['principal_investigator_name'] = row[1].split("//")[0].strip(' "')
# 
#                     #principal investigator institution
#                     if row[0] == "*PRINCIPAL INVESTIGATOR AFFILIATION // AFFILIATION DU CHERCHEUR PRINCIPAL":
#                         self.metadata['principal_investigator_institution'] = row[1].split("//")[0].strip(' "')
# 
#                     #network_qa_details + data_level
#                     if row[0] == "'*QUALITY ASSESSMENT LEVEL // NIVEAU D\x92ÉVALUATION DE LA QUALITÉ'": 
#                         
#                     #network_sampling_details
#                     
#                 #in station metadata section?
#                 if station_metadata_section = True:
#         
#                 #in data section?
#                 if data_section:
#                             
# 
# 
# 
#                 #start processing metadata - taking care with french accents
#                 #find start of metadata chunk
#                 #use try/except for blank lines
#                 try:
#                     if (row[0] =='*TABLE NAME // NOM DE LA TABLE') & (row[1] == 'Site information'):
#                         metadata_chunk = True
#                 except:
#                     pass
# 
#                 if metadata_chunk:
#                     #get names of metadata variables 
#                     if (row[0] =='*TABLE COLUMN NAME--SHORT FORM // NOM DE COLONNE--FORMAT COURT'):
#                         meta_variable_names = row
#             
#                     #get ready to ready metadata for each site
#                     if (crossnetwork_processing_functions.remove_accents(row[0]) == '*TABLE DATA BEGINS // DEBUT DES DONNEES DU TABLEAU'):
#                         read_metadata = True
#                         continue
#                         
#                     #stop processing metadata
#                     if (crossnetwork_processing_functions.remove_accents(row[0]) == '*TABLE DATA ENDS // FIN DES DONNEES DU TABLEAU'):
#                         read_metadata = False
#                         metadata_chunk = False
#                 
#                     #read metadata line
#                     if read_metadata:
#                     
#                         #check each string in metadata to remove accents
#                         #strip leading and trailing whitespace
#                         #change all fields to lower case
#                         row = [crossnetwork_processing_functions.remove_accents(i).strip().lower() for i in row]
#                         
#                         #put metadata into lists
#                         metadata[metadata_headers.index('reference')].append(row[meta_variable_names.index('SiteID')])
#                         metadata[metadata_headers.index('site_name')].append(row[meta_variable_names.index('SiteName')].split('//')[0].strip())
#                         metadata[metadata_headers.index('country')].append(row[meta_variable_names.index('CountryCode')].split('(')[-1].split(')')[0].strip())
#                         metadata[metadata_headers.index('province_state_code')].append(row[meta_variable_names.index('ProvinceStateCode')])
#                         metadata[metadata_headers.index('latitude')].append(row[meta_variable_names.index('Latitude_deg')])
#                         metadata[metadata_headers.index('longitude')].append(row[meta_variable_names.index('Longitude_deg')])
#                         metadata[metadata_headers.index('altitude')].append(row[meta_variable_names.index('GroundElevAMSL_m')])
#                         metadata[metadata_headers.index('datum')].append(row[meta_variable_names.index('LatLonRefDatum')].split('(')[0].strip())
#                         metadata[metadata_headers.index('timezone')].append(row[meta_variable_names.index('Timezone')])
#                         metadata[metadata_headers.index('land_use')].append(row[meta_variable_names.index('SiteLandUse')].split('//')[0].strip())
#                         metadata[metadata_headers.index('sampling_height')].append(row[meta_variable_names.index('SamplingHeightAG')])
#                         metadata[metadata_headers.index('meta_year')].append(process_year)
#                         
#                         #process sampling heights into workable format, as given as time ranges across time record
#                     
#                         #remove whitespace, remove character 'm', add split string to identify changing metadata, convert back to unicode 
#                         metadata[metadata_headers.index('sampling_height')][-1] = unicode(metadata[metadata_headers.index('sampling_height')][-1].replace(" ", "").replace("m","").replace(",","[CHANGINGMETASPLIT]").replace(";","[CHANGINGMETASPLIT]"))
#                         #change dates given into YYYYMMDDHHMM format
#                         sampling_height_split = metadata[metadata_headers.index('sampling_height')][-1].split('[CHANGINGMETASPLIT]')
#                         amended_date_range = []
#                         #iterate through different sampling heights for a site
#                         for i,var in enumerate(sampling_height_split):
#                             date_value_split = var.split(':')
#                             #determine if different data ranges are given for a site (otherwise just a number is given)
#                             if len(date_value_split) != 1:
#                                 value = date_value_split[-1]
#                                 date_split = date_value_split[0].split('-')
#                                 amended_dates = []
#                                 #iterate through different dates provided
#                                 for x,var2 in enumerate(date_split):
#                                     if var2 != 'onwards':
#                                         if len(var2) == 4:
#                                             dd = datetime.datetime.strptime(var2,'%Y')
#                                             #if date is the closing year, then add year to have no gaps in datetimes
#                                             if x == 1:
#                                                 dd = dd + relativedelta(years=1)
#                                         else:
#                                             dd = datetime.datetime.strptime(var2,'%d%b%Y')
#                                             #if date is the closing year, then add day to have no gaps in datetimes
#                                             if x == 1:
#                                                 dd = dd + relativedelta(days=1)
#                                     else:
#                                         dd = 'onwards'
#                                     #convert datetimes to output format (YYYYMMDDHHMM)
#                                     try:
#                                         dd = dd.strftime('%Y%m%d%H%M')
#                                     except:
#                                         pass
#                                     amended_dates.append(dd)
#                                 amended_date_range.append(amended_dates[0]+'-'+amended_dates[-1]+':'+value)
#                         
#                         #if date ranges are given then put together in required format (YYYYMMDDHHMM-YYYYMMDDHHMM:value[CHANGINGMETASPLIT]YYYYMMDDHHMM-onwards:value)
#                         if len(amended_date_range) != 0:
#                             new_sampling_height = ''
#                             for n in range(len(amended_date_range)):
#                                 if n != (len(amended_date_range)-1):
#                                     new_sampling_height = new_sampling_height+amended_date_range[n]+'[CHANGINGMETASPLIT]'
#                                 else:
#                                     new_sampling_height = new_sampling_height+amended_date_range[n]
#                             metadata[metadata_headers.index('sampling_height')][-1] = unicode(new_sampling_height)
#                         
#                 #start processing data 
#                 #find start of data chunk
#                 #use try/except for blank lines   
#                 try:
#                     if (row[0] =='*TABLE NAME // NOM DE LA TABLE') & (row[1] == 'CAPMoN_Ozone'):
#                         data_chunk = True
#                 except:
#                     pass
#            
#                 if data_chunk:
#                     #get names of metadata variables 
#                     if (row[0] =='*TABLE COLUMN NAME--SHORT FORM // NOM DE COLONNE--FORMAT COURT'):
#                         data_variable_names = row
#                     
#                     #get ready to ready data for each site
#                     if (crossnetwork_processing_functions.remove_accents(row[0]) == '*TABLE DATA BEGINS // DEBUT DES DONNEES DU TABLEAU'):
#                         read_data = True
#                         continue
#                     
#                     #stop processing data
#                     if (crossnetwork_processing_functions.remove_accents(row[0]) == '*TABLE DATA ENDS // FIN DES DONNEES DU TABLEAU'):
#                         read_data = False
#                         data_chunk = False
#                         
#                     #read data line
#                     if read_data:
#                         #put data into lists
#                         data[data_headers.index('reference')].append(row[data_variable_names.index('SiteID')])
#                         data[data_headers.index('start_date_UTC')].append(row[data_variable_names.index('DateStartUTC')])
#                         data[data_headers.index('start_time_UTC')].append(row[data_variable_names.index('TimeStartUTC')])
#                         data[data_headers.index('end_date_UTC')].append(row[data_variable_names.index('DateEndUTC')])
#                         data[data_headers.index('end_time_UTC')].append(row[data_variable_names.index('TimeEndUTC')])
#                         data[data_headers.index('concentration')].append(row[data_variable_names.index('O3_ppbv')])
#                         data[data_headers.index('flag')].append(row[data_variable_names.index('O3_ppbv_Flag')])
#                         data[data_headers.index('lod')].append(row[data_variable_names.index('O3_ppbv_DetL')])
# 
#     self.metadata,self.metadata_headers,self.data,self.data_headers = metadata,metadata_headers,data,data_headers                      
        

class ParseNASAAmesFile:

    '''class that handles processing of a nasa ames file'''

    def __init__(self,filename,station_reference,run_details):
        
        self.filename = filename
        self.station_reference = station_reference
        self.run_details = run_details

        #core header meta
        self.n_header = ''          #number of header lines
        self.ffi = ''               #file format index
                                    #1st digit indicates the number of independent variables (dimensions)
                                    #2nd digit indicates number of necessary auxiliary variable
                                    #Remaining differences cover the way the independent variables increment and the nature (numeric or character) of the variables
        self.originator_name = ''   #data originator name
        self.originator_affil = ''  #data originator affiliation
        self.provider_name = ''     #data provider name
        self.campaign_name = ''     #campaign name
        self.volume_number = ''     #current volume number (dated field - should be 1)
        self.n_volumes = ''         #number of volumes in dataset (dated field - should be 1)
        self.date_start = ''        #date for which record starts
        self.file_creation = ''     #date of file creation


        #independent variable meta (e.g. time)
        self.independent_interval = ''  #independent variable interval (if changes should be zero)
        self.independent_name = ''      #independent variable name (e.g. days from file reference point)


        #primary variable meta (quantities that have been observed or derived)
        self.n_primary = ''               #number of primary variables
        self.primary_scale_factors = ''   #scale factors for each of primary variables
        self.primary_missing_values = ''  #value imposed for missing data for each primary variable
        self.primary_names = []           #primary variable names (e.g. ozone)
        self.primary_units = []           #primary variable units
        self.primary_details = []         #extra details about variables (e.g. limit of detection)

        #auxiliary variable meta
        self.n_auxiliary = ''    #number of auxiliary variables

        #comments meta
        self.n_comments = ''     #number of lines of comments
        self.comments = []       #extra meta comments about data. Can also include column headings

        #important line counts
        self.endline_primary_count = -999
        self.start_auxiliary = -999
        self.start_comments = -999
        self.start_data = -999
        self.row_count = 0

    def read_header(self):

        with open(self.filename) as f:
            self.content = f.readlines()
            self.content = [x.strip('\n') for x in self.content]

            for row in self.content:

                #--read core header meta--
                if self.row_count == 0:
                    line_split = row.split(' ')
                    self.n_header,self.ffi = line_split[0],line_split[1]

                elif self.row_count == 1:
                    self.originator_name = row

                elif self.row_count == 2:
                    self.originator_affil = row

                elif self.row_count == 3:
                    self.provider_name = row

                elif self.row_count == 4:
                    self.campaign_name = row

                elif self.row_count == 5:
                    line_split = row.split(' ')
                    self.volume_name, self.n_volumes = line_split[0],line_split[1]

                elif self.row_count == 6:
                    line_split = row.split(' ')
                    self.date_start, self.file_creation = line_split[0]+line_split[1]+line_split[2], line_split[3]+line_split[4]+line_split[5]

                #--read independent variable meta--
                elif self.row_count == 7:
                    self.independent_interval = row

                elif self.row_count == 8:
                    self.independent_name = row

                #--read primary variable meta--
                elif self.row_count == 9:
                    self.n_primary = int(row)
                    self.endline_primary_count = 11+self.n_primary
                    self.start_auxiliary = self.endline_primary_count+1

                elif self.row_count == 10:
                    self.primary_scale_factors = row.split(' ')

                elif self.row_count == 11:
                    self.primary_missing_values = row.split(' ')

                elif (self.row_count > 11) & (self.row_count <= self.endline_primary_count):
                    #handle line splitting with quotes (with commas inside)
                    line_split = []
                    split_by_quotes = row.split('"')
                    for split_ii, split_element in enumerate(split_by_quotes):
                        if (split_ii % 2) == 0:
                            line_split = np.append(line_split, split_element.split(','))
                        else:
                            if split_element[:7] == 'Comment':
                                new_str = 'Comment="{}"'.format(split_element[8:])
                            else: 
                                new_str = '{}"{}"'.format(line_split[-1], split_element)
                                line_split = line_split[:-1]
                            line_split = np.append(line_split, new_str)
                    line_split = [element.strip() for element in line_split]
                    line_split = [element for element in line_split if element != '']
                    if len(line_split) == 0:
                        line_split = [row]

                    #only read line if first value is one of accepted primary variable names
                    #otherwise line is appended to primary details - this deals with badly formatted files
                    #also add 1 to auxiliary start count - so don't start reading too soon

                    #create a list of all potentially valid primary variables
                    valid_primary_variables = ['end_time of measurement','numflag']
                    for param in list(standard_parameters.keys()):
                        ebas_parameter_name = standard_parameters[param]['ebas_parameter_name']
                        #only process variables with a valid EBAS parameter name (i.e. not empty string)
                        if ebas_parameter_name != '':
                            valid_primary_variables=np.append(valid_primary_variables,ebas_parameter_name)
                            #create numflag variable for all parameters
                            valid_primary_variables=np.append(valid_primary_variables,'numflag {}'.format(ebas_parameter_name))                    

                    if line_split[0] in valid_primary_variables:
                        self.primary_names.append(line_split[0])
                        self.primary_units.append(line_split[1])
                        if len(line_split) > 2:
                            self.primary_details.append(line_split[2:])
                        else:
                            self.primary_details.append([])
                    else:
                        self.primary_details[-1] = self.primary_details[-1]+line_split
                        self.endline_primary_count+=1
                        self.start_auxiliary+=1

                #--read auxiliary variable meta--
                #first checks if row is a valid number or relates to the previous primary details not correctly formatted
                #if so then append to primary details
                elif self.row_count == self.start_auxiliary:
                    try:
                        self.n_auxiliary = int(row)
                        #if is not 0 then break processing and panic - as have never seen this be used and don't know format
                        if self.n_auxiliary != 0:
                            self.run_details['log_file_str'] += 'Using auxilary variables. {} of them! How Strange. Now need to code for this. In file: {} .Stopping processing.'.format(self.n_auxiliary, self.filename)
                            create_GHOST_stage_logfile(2,self.run_details)
                        self.start_comments = self.start_auxiliary+1
                    except:
                        line_split = row.split(',')
                        self.primary_details[-1] = np.append(self.primary_details[-1],line_split)
                        self.endline_primary_count+=1
                        self.start_auxiliary+=1

                #--read comments meta--
                elif self.row_count == self.start_comments:
                    self.n_comments = int(row)
                    self.end_comments = self.start_comments+self.n_comments
                    self.start_data = self.end_comments+1

                elif (self.row_count > self.start_comments) & (self.row_count <= self.end_comments):
                    self.comments.append(row)

                #manually iterate row count
                self.row_count+=1

    def read_data(self):
        #cut file read to just be data columns
        data = self.content[self.start_data:]

        #make list of lists to store data columns (1 independent variable + primary variables)
        self.structured_data = [[] for _ in range(self.n_primary+1)]

        #iterate through lines, appending column data into a list of lists
        for line in data:
            line_split = line.split()
            for n,var in enumerate(line_split):
                self.structured_data[n].append(var)


"""A dictionary, where key is the ISD identifier in ASCII and value
is a list that contains [variable_name, number_of_chars]. The variable
name is as in ISD format. The number of chars indicate the number of
characters data (without considering the 3 chars of the identifier"""


isd_map = {'AA1': ['LIQUID-PRECIPITATION', 8],
           'AA2': ['LIQUID-PRECIPITATION', 8],
           'AA3': ['LIQUID-PRECIPITATION', 8],
           'AA4': ['LIQUID-PRECIPITATION', 8],
           'AB1': ['LIQUID-PRECIPITATION MONTHLY TOTAL', 7],
           'AC1': ['PRECIPITATION-OBSERVATION-HISTORY', 3],
           'AD1': ['LIQUID-PRECIPITATION GREATEST AMOUNT IN 24 HOURS, FOR THE MONTH', 19],
           'AE1': ['LIQUID-PRECIPITATION, NUMBER OF DAYS WITH SPECIFIC AMOUNTS, FOR THE MONTH', 12],
           'AG1': ['PRECIPITATION-ESTIMATED-OBSERVATION', 4],
           'AH1': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH2': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH3': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH4': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH5': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AH6': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI1': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI2': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI3': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI4': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI5': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AI6': ['LIQUID-PRECIPITATION MAXIMUM SHORT DURATION', 15],
           'AJ1': ['SNOW-DEPTH', 14],
           'AK1': ['SNOW-DEPTH-GREATEST-ON-GROUND', 12],
           'AL1': ['SNOW-ACCUMULATION', 7],
           'AL2': ['SNOW-ACCUMULATION', 7],
           'AL3': ['SNOW-ACCUMULATION', 7],
           'AL4': ['SNOW-ACCUMULATION', 7],
           'AM1': ['SNOW-ACCUMULATION-GREATEST-IN-24H', 18],
           'AN1': ['SNOW-ACCUMULATION-FOR-MONTH', 9],
           'AO1': ['LIQUID-PRECIPITATION', 8],
           'AO2': ['LIQUID-PRECIPITATION', 8],
           'AO3': ['LIQUID-PRECIPITATION', 8],
           'AO4': ['LIQUID-PRECIPITATION', 8],
           'AP1': ['15MIN-LIQUID-PRECIP', 6],
           'AP2': ['15MIN-LIQUID-PRECIP', 6],
           'AP3': ['15MIN-LIQUID-PRECIP', 6],
           'AP4': ['15MIN-LIQUID-PRECIP', 6],
           'AT1': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],             # Weather Occurrence Data
           'AT2': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT3': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT4': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT5': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT6': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT7': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AT8': ['DAILY-PRESENT-WEATHER-OBSERVATION', 9],
           'AU1': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU2': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU3': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU4': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU5': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU6': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU7': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU8': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AU9': ['PRESENT-WEATHER-OBSERVATION', 8],
           'AW1': ['PRESENT-WEATHER-OBSERVATION-AUTOMATED', 3],
           'AW2': ['PRESENT-WEATHER-OBSERVATION-AUTOMATED', 3],
           'AW3': ['PRESENT-WEATHER-OBSERVATION-AUTOMATED', 3],
           'AX1': ['PAST-WEATHER-OBSERVATION', 6],
           'AX2': ['PAST-WEATHER-OBSERVATION', 6],
           'AX3': ['PAST-WEATHER-OBSERVATION', 6],
           'AX4': ['PAST-WEATHER-OBSERVATION', 6],
           'AX5': ['PAST-WEATHER-OBSERVATION', 6],
           'AX6': ['PAST-WEATHER-OBSERVATION', 6],
           'AY1': ['PAST-WEATHER-OBSERVATION manual', 5],
           'AY2': ['PAST-WEATHER-OBSERVATION manual', 5],
           'AZ1': ['PAST-WEATHER-OBSERVATION automated', 5],
           'AZ2': ['PAST-WEATHER-OBSERVATION automated', 5],
           'CB1': ['Subhourly Liquid Precipitation Section', 10],       # Climate Reference Network Unique Data
           'CB2': ['Subhourly Liquid Precipitation Section', 10],
           'CF1': ['Hourly Fan Speed Section', 6],
           'CF2': ['Hourly Fan Speed Section', 6],
           'CF3': ['Hourly Fan Speed Section', 6],
           'CG1': ['PRIMARY-SENSOR', 8],
           'CG2': ['PRIMARY-SENSOR', 8],
           'CG3': ['PRIMARY-SENSOR', 8],
           'CH1': ['HOURLY-RELATIVE-HUMIDITY', 15],
           'CH2': ['HOURLY-RELATIVE-HUMIDITY', 15],
           'CI1': ['RELATIVE-HUMIDITY-TEMPERATURE', 28],
           'CN1': ['HOURLY-BATTERY-VOLTAGE', 18],
           'CN2': ['HOURLY-DIAGNOSTIC', 18],
           'CN3': ['SECONDARY-HOURLY-DIAGNOSTIC', 16],
           'CN4': ['SECONDARY-HOURLY-DIAGNOSTIC', 19],
           'CO1': ['US-NETWORK-METADATA', 5],                           # Network Metadata
           'CO2': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO3': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO4': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO5': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO6': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO7': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO8': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CO9': ['US-COOP-NETWORK-TIME-OFFSET', 8],
           'CR1': ['CONTROL-SECTION', 7],
           'CT1': ['SUBHOURLY-TEMPERATURE', 7],
           'CT2': ['SUBHOURLY-TEMPERATURE', 7],
           'CT3': ['SUBHOURLY-TEMPERATURE', 7],
           'CU1': ['HOURLY-TEMPERATURE', 13],
           'CU2': ['HOURLY-TEMPERATURE', 13],
           'CU3': ['HOURLY-TEMPERATURE', 13],
           'CV1': ['HOURLY-TEMPERATURE-EXTREME', 26],
           'CV2': ['HOURLY-TEMPERATURE-EXTREME', 26],
           'CV3': ['HOURLY-TEMPERATURE-EXTREME', 26],
           'CW1': ['SUBHOURLY-WETNESS', 14],
           'CX1': ['HOURLY-GEONOR-VIBRATING-WIRE', 26],
           'CX2': ['HOURLY-GEONOR-VIBRATING-WIRE', 26],
           'CX3': ['HOURLY-GEONOR-VIBRATING-WIRE', 26],
           'ED1': ['RUNWAY-VISUAL-RANGE', 8],                           # Runway Visual Range Data
           'GA1': ['SKY-COVER-LAYER', 13],                              # Cloud and Solar Data
           'GA2': ['SKY-COVER-LAYER', 13],
           'GA3': ['SKY-COVER-LAYER', 13],
           'GA4': ['SKY-COVER-LAYER', 13],
           'GA5': ['SKY-COVER-LAYER', 13],
           'GA6': ['SKY-COVER-LAYER', 13],
           'GD1': ['SKY-COVER-SUMMATION', 12],
           'GD2': ['SKY-COVER-SUMMATION', 12],
           'GD3': ['SKY-COVER-SUMMATION', 12],
           'GD4': ['SKY-COVER-SUMMATION', 12],
           'GD5': ['SKY-COVER-SUMMATION', 12],
           'GD6': ['SKY-COVER-SUMMATION', 12],
           'GE1': ['SKY-CONDITION', 19],
           'GF1': ['SKY-CONDITION', 23],
           'GG1': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG2': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG3': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG4': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG5': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GG6': ['BELOW-STATION-CLOUD-LAYER', 15],
           'GH1': ['HOURLY-SOLAR-RADIATION', 28],
           'GJ1': ['SUNSHINE', 5],
           'GK1': ['SUNSHINE-OBSERVATION', 4],
           'GL1': ['SUSHINE-OBSERVATION-FOR-MONTH', 6],
           'GM1': ['SOLAR-IRRADIANCE', 30],
           'GN1': ['SOLAR-RADIATION', 28],
           'GO1': ['NET-SOLAR-RADIATION', 19],
           'GP1': ['MODELED-SOLAR-IRRADIANCE', 31],
           'GQ1': ['HOURLY-SOLAR-ANGLE', 14],
           'GR1': ['HOURLY-EXTRATERRESTRIAL-RADIATION', 14],
           '???': ['HAIL', 4],                                          # Hail Data - Missing ID in the Doc
           'IA1': ['GROUND-SURFACE', 3],                                # Ground Surface Data
           'IA2': ['GROUND-SURFACE', 9],
           'IB1': ['HOURLY-SURFACE', 27],
           'IB2': ['HOURLY-SURFACE', 13],
           'IC1': ['GROUND-SURFACE', 25],
           'KA1': ['EXTREME-AIR-TEMPERATURE', 10],                      # Temperature Data
           'KA2': ['EXTREME-AIR-TEMPERATURE', 10],
           'KA3': ['EXTREME-AIR-TEMPERATURE', 10],
           'KA4': ['EXTREME-AIR-TEMPERATURE', 10],
           'KB1': ['AVERAGE-AIR-TEMPERATUER', 10],
           'KB2': ['AVERAGE-AIR-TEMPERATUER', 10],
           'KB3': ['AVERAGE-AIR-TEMPERATUER', 10],
           'KC1': ['EXTREME-AIR-TEMP-FOR-MONTH', 14],
           'KC2': ['EXTREME-AIR-TEMP-FOR-MONTH', 14],
           'KD1': ['HEATING-COOLING-DEGREE-DAYS', 9],
           'KD2': ['HEATING-COOLING-DEGREE-DAYS', 9],
           'KE1': ['EXTREME-TEMPS-EXCEEDING-CRITERIA', 12],
           'KF1': ['HOURLY-CALCULATED-TEMP', 6],
           'KG1': ['AVERAGE-DEW_POINT', 11],
           'KG2': ['AVERAGE-DEW_POINT', 11],
           'MA1': ['ATMOSPHERIC-PRESSURE-OBSERVATION', 12],             # Pressure Data
           'MD1': ['ATMOSPHERIC-PRESSURE-CHANGE', 11],
           'ME1': ['GEOPOTENTIAL-HEIGHT', 6],
           'MF1': ['ATMOSPHERIC-PRESSURE-STP-SLP', 12],
           'MG1': ['ATMOSPHERIC-PRESSURE', 12],
           'MH1': ['ATMOSPHERIC-PRESSURE-FOR-MONTH', 12],
           'MK1': ['ATMOSPHERIC-PRESSURE-FOR-MONTH', 24],
           'MV1': ['PRESENT-WEATHER-IN-VICINITY', 3],                   # Weather Occurrence Data (Again?)
           'MV2': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV3': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV4': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV5': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV6': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MV7': ['PRESENT-WEATHER-IN-VICINITY', 3],
           'MW1': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW2': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW3': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW4': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW5': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW6': ['PRESENT-WEATHER-OBSERVATION', 3],
           'MW7': ['PRESENT-WEATHER-OBSERVATION', 3],
           'OA1': ['SUPPLEMENTARY-WIND-OBSERVATION', 8],                # Wind Data
           'OA2': ['SUPPLEMENTARY-WIND-OBSERVATION', 8],
           'OA3': ['SUPPLEMENTARY-WIND-OBSERVATION', 8],
           'OB1': ['Hourly/Sub-Hourly Wind Section', 28],
           'OB2': ['Hourly/Sub-Hourly Wind Section', 28],
           'OC1': ['WIND-GUST-OBSERVATION', 5],
           'OD1': ['SUPPLEMENTARY-WIND-OBSERVATION', 11],
           'OD2': ['SUPPLEMENTARY-WIND-OBSERVATION', 11],
           'OD3': ['SUPPLEMENTARY-WIND-OBSERVATION', 11],
           'RH1': ['RELATIVE HUMIDITY', 9],
           'RH2': ['RELATIVE HUMIDITY', 9],
           'RH3': ['RELATIVE HUMIDITY', 9],
           'SA1': ['SEA-SURFACE-TEMPERATURE', 5],                       # Sea Surface Temperature Data
           'ST1': ['SOIL-TEMPERATURE', 17],                             # Soil Temperature Data
           'UA1': ['WAVE-MEASUREMENT', 10],                             # Marine Data
           'UG1': ['WAVE-MEASUREMENT-PRIMARY', 9],
           'UG2': ['WAVE-MEASUREMENT-SECONDARY', 9],
           'WA1': ['PLATFORM-ICE-ACCRETION', 6],
           'WD1': ['WATER-SURFACE-ICE', 20],
           'WG1': ['WATER-SURFACE-ICE-HISTORICAL', 11],
           'WJ1': ['WATER-LEVEL-OBSERVATION', 19]}

class ISDLineParser:

    def __init__(self, filename, row):

        self.filename = filename
        self.prec_period = None
        self.add_offset = 105  # always fixed
        self.section_chars = 3
        self.missing = 9999

        self.varmap = isd_map

        self.sampling_height = None  # to be assigned in NCDC_ISD_Processor
        self.measurement_altitude = None  # to be assigned in NCDC_ISD_Processor
        self.station_elevation = None  # to be assigned in NCDC_ISD_Processor

        # Include all variables that should appear to every file in __init__ for readability
        # Initialize all to None, then call parsing methods

        self.num_char = None

        # Control section attributes
        self.station_idu = None
        self.station_idn = None
        self.date_ascii = None
        self.time_ascii = None
        self.data_source_flag = None
        self.latitude = None
        self.longitude = None
        self.code = None
        self.elevation = None
        self.call_letter = None
        self.qc_name = None

        # Mandatory section attributes
        # wind-obdervation
        self.direction_angle = None
        self.direction_angle_qc = None
        self.type_code = None
        self.speed = None
        self.speed_qc = None

        # sky-condition
        self.ceiling = None
        self.ceiling_qc = None
        self.ceiling_dc = None

        # visibility
        self.distance = None
        self.distance_qc = None
        self.variability = None
        self.variability_qc = None

        # air-temperature"
        self.air_temp = None
        self.air_temp_qc = None

        # air-temperature dew point"
        self.dpoint_temp = None
        self.dpoint_temp_qc = None

        # atmospheric pressure
        self.sea_level_pressure = None
        self.sea_level_pressure_qc = None

        # Call the first method to start the initilization of the attributes
        # self.open_file()
        self.parse_file_line(row)

        # Remarks Section
        self.rem_content = None

        # Original Observation Section
        self.has_qnn = False

    def parse_file_line(self, line):
        # First read how many chars the line has
        self.num_char = int(line[0:4]) + self.add_offset
        # Read control and mandatory section
        self.parse_control_section(line)
        self.parse_mandatory_section(line)

        # Additional Data Section
        additional = line[105:108]
        pos = 105
        if additional == "ADD":
            # try to read
            pos += self.section_chars
            while pos < self.num_char:
                try:
                    read = self.parse_add_section(line, pos)
                    pos += read
                except Exception as e:
                    # reached REM
                    # logging.exception(e)
                    break
        while pos < self.num_char:
            if line[pos:pos + self.section_chars] == 'REM':
                try:
                    rem_text = self.parse_rem_section(line[pos:self.num_char])
                    pos += len(rem_text)
                except Exception as e:
                    break
            else:
                break
        if line[pos:pos + self.section_chars] == 'EQD':
            # while pos < num_char:
            try:
                read = self.parse_eqd_section(line[pos:self.num_char])
                pos += read
            except Exception as e:
                print(line[pos:pos + self.section_chars])  # change to logging
        if line[pos:pos + self.section_chars] == 'QNN':
            self.parse_qnn_section()
            print(pos, self.num_char)
        # if pos < self.num_char:


    def parse_control_section(self, line):
        """Function to read the Control Data Section"""

        # scaling factors
        scale_lats = 1000

        self.station_idu = line[4:10]  # station, USAF catalog identifier
        self.station_idn = line[10:15]  # station, NCEI catalog identifier
        self.date_ascii = line[15:23]  # date of observation
        self.time_ascii = line[23:27]  # time on UTC
        self.datetime_ascii = line[15:27]
        self.data_source_flag = line[17]  #
        self.latitude = float(line[28:34]) / scale_lats
        if self.latitude == 99.999:
            self.latitude = np.NaN
        self.longitude = float(line[34:41]) / scale_lats  # 000000 to 179999 are signed negative
        if self.longitude == 999.999:
            self.longitude = np.NaN
        self.code = line[41:46]  # type of geophysical surface observation
        self.elevation = float(line[46:51])  # relative to MSL
        if self.elevation == 9999:
            self.elevation = np.NaN
        # self.call_letter = line[51 - 56]  # no idea
        # self.qc_name = line[56:60]  # quality control process name

    def parse_mandatory_section(self, line):
        """Function to read the Mandatory Data Section"""

        # scaling factors
        scale_speed = 10
        scale_temp = 10
        scale_slp = 10
        
        # constant to convert to kelvin
        kelvin = 273.15

        "Wind-Observation"
        self.direction_angle = float(line[60:63])
        self.direction_angle_qc = line[63]
        if self.direction_angle == 999:
            self.direction_angle_qc = '3'
        self.speed = float(line[65:69]) / scale_speed
        self.speed_qc = line[69]
        if self.speed == 999.9:
            self.speed_qc = '3'

        "Sky-Condition-Observation"
        self.ceiling = float(line[70:75])
        self.ceiling_qc = line[75]
        if self.ceiling == 99999:
            self.ceiling_qc = '3'
        if self.ceiling == 22000:
            self.ceiling_qc = 'CHM'

        "Visibility-Observation"
        self.distance = float(line[78:84])
        self.distance_qc = line[84]
        if self.distance == 999999 or self.distance_qc == 9999:
            self.distance_qc = '3'
        if self.distance == 160000:
            self.distance_qc = 'VIM'

        "Air-Temperature-Observation"
        self.air_temp = (float(line[87:92]) / scale_temp) + kelvin
        self.air_temp_qc = line[92]
        if self.air_temp == 999.9:
            self.air_temp_qc = '3'

        "Air-Temperature-Observation dew point"
        self.dpoint_temp = (float(line[93:98]) / scale_temp) + kelvin
        self.dpoint_temp_qc = line[98]
        if self.dpoint_temp == 999.9:
            self.dpoint_temp_qc = '3'

        "Atmospheric-Pressure-Observation"
        self.sea_level_pressure = float(line[99:104]) / scale_slp
        self.sea_level_pressure_qc = line[104]
        if self.sea_level_pressure == 9999.9:
            self.sea_level_pressure_qc = '3'

    def parse_add_section(self, line, position):
        """ Function to parse the Additional Data Section"""

        identifier = line[position:position + self.section_chars]

        # where the section of data start and end according to the id
        start_block = position + self.section_chars
        end_block = start_block + isd_map[identifier][1]
        block = line[start_block:end_block]

        # the data from the additional section
        # call function find_add_id
        self.find_add_id(identifier, block)

        # send back how many characters were read
        read = isd_map[identifier][1] + self.section_chars
        return read

    def find_add_id(self, identifier, block, print_info=False, print_block=False):
        """ According to the input identifier, the function reads respectively
        the variables under the identifier, following the isd format document.
        Unfortunately, this is by far a not elegant way. Suggestions are welcome."""

        map_value = 0  # position of the name in isd_map
        # scale factor - similar for all values thankfully
        scale_factor = 10
        kelvin = 273.15

        if identifier in ('AA1', 'AA2', 'AA3', 'AA4', 'AO1', 'AO2', 'AO3', 'AO4'):
            if identifier[-1] in ['2', '3', '4']:
                return
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that includes precipitation depth
            self.prec_period = float(block[0:2]) * 60  # because we want minutes
            if self.prec_period == 0:
                return
            self.prec_depth = float(block[2:6]) / scale_factor
            self.prec_depth_qc = block[7]
            if self.prec_depth == 999.9:
                self.prec_depth_qc = '3'

        elif identifier == 'AL1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that includes snow accumulation
            self.snow_acc_period = float(block[0:2])  * 60  # because we want minutes
            if self.snow_acc_period == 0:
                return
            self.snow_acc = float(block[2:5])
            self.snow_acc_qc = block[6]
            if self.snow_acc == 999:
                self.snow_acc_qc = '3'

        elif identifier == 'AJ1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that includes snow depth
            self.si = float(block[0:4])
            self.si_qc = block[5]
            if self.si == 9999:
                self.si_qc = '3'

        elif identifier in ('CH1', 'CH2'):
            if print_info:
                print(isd_map[identifier][map_value])
            self.avg_rh_period = float(block[0:2])  # already in minutes
            self.avg_rh = float(block[9:13]) / scale_factor
            self.avg_rh_qc = block[13]
            if self.avg_rh == 999.9:
                self.avg_rh_qc = '3'

        elif identifier == 'MA1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that includes station pressure rate
            self.station_p = float(block[6:11]) / scale_factor
            self.station_p_qc = block[11]
            if self.station_p == 9999.9:
                self.station_p_qc = '3'

        elif identifier in ('GA1', 'GA2', 'GA3', 'GA4', 'GA5', 'GA6', 'GD1'):
            # if we have already read a value for cloud coverage in the same line,
            # then keep the field with maximum value
            if identifier == 'GA1':
                self.cloud = int(block[0:2])
                self.cloud_qc = block[2]
                # keep flag to know that line has GA
                self.has_ga = True
                if self.cloud == 99:
                    self.cloud_qc = '3'
            elif identifier in ('GA2', 'GA3', 'GA4', 'GA5', 'GA6'):
                new_value = int(block[0:2])
                if new_value != 99 and new_value > self.cloud:
                    self.cloud = int(block[0:2])
                    self.cloud_qc = block[2]
                    self.has_ga = True
            elif identifier == 'GD1':
                # First of all, read the GD parameter
                # read part that includes sky coverage summation code
                self.cloud_gd = int(block[1:3])
                self.cloud_gd_qc = block[3]
                # if code 2 is missing, take code 1 and convert to oktas
                if self.cloud_gd == 99 or int(self.cloud_cov) > 8:
                    code1 = int(block[0])
                    # convert code to oktas
                    if code1 == 0:
                        self.cloud_gd = 0
                    elif code1 == 1:
                        self.cloud_gd = 2
                    elif code1 == 2:
                        self.cloud_gd = 4
                    elif code1 == 3:
                        self.cloud_gd = 6
                    elif code1 == 4:
                        self.cloud_gd = 8
                    else:
                        self.cloud_gd = 99
                        self.cloud_gd_qc = '3'
                # Now that we have the code, check what if GA exists, using the flag
                if self.cloud:
                    # if it has ga, check if ga == 99
                    if self.cloud == 99:
                        # then replace with the value we keep from gd
                        self.cloud = self.cloud_gd
                        self.cloud_qc = self.cloud_gd_qc
                elif not self.has_ga:
                    self.cloud = self.cloud_gd
                    self.cloud_qc = self.cloud_gd_qc
            # at this point we should have the final value that we keep for cloud
            # so following, we can compute the cloud fraction
            if self.cloud == 99:
                self.cloud_frac = 999
                self.cloud_frac_qc = '3'
            elif self.cloud == 1:
                self.cloud_frac = 12.5
                self.cloud_frac_qc = '0'
            elif self.cloud == 2:
                self.cloud_frac = 25
                self.cloud_frac_qc = '0'
            elif self.cloud == 3:
                self.cloud_frac = 37.5
                self.cloud_frac_qc = '0'
            elif self.cloud == 4:
                self.cloud_frac = 50
                self.cloud_frac_qc = '0'
            elif self.cloud == 5:
                self.cloud_frac = 62.5
                self.cloud_frac_qc = '0'
            elif self.cloud == 6:
                self.cloud_frac = 75
                self.cloud_frac_qc = '0'
            elif self.cloud == 7:
                self.cloud_frac = 87.5
                self.cloud_frac_qc = '0'
            elif self.cloud == 8:
                self.cloud_frac = 100
                self.cloud_frac_qc = '0'

        elif identifier == 'SA1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that contains sea surface temperature
            self.sst = (float(block[0:4]) / scale_factor) + kelvin
            self.sst_qc = block[4]
            if self.sst == 99.9:
                self.sst_qc = '3'

        elif identifier == 'ST1':
            if print_info:
                print(isd_map[identifier][map_value])
            # read part that contains soil temperature data
            self.soil_temp = (float(block[1:6]) / scale_factor) + kelvin
            self.soil_temp_qc = block[6]
            if self.soil_temp == 999.9:
                self.soil_temp_depth_qc = '3'
            self.soil_temp_depth = float(block[7:11]) / scale_factor
            self.soil_temp_depth_qc = block[11]
            if self.soil_temp_depth == 999.9:
                self.soil_temp_depth_qc = '3'

       # Read Weather Condition Codes (includes dust-related codes)
        elif identifier in ('AT1', 'AT2', 'AT3', 'AT4', 'AT5', 'AT6', 'AT7', 'AT8'):
            if print_info:
                print(isd_map[identifier][map_value])
            # discard flag if its qa code is defined as invalid
            if block[-1:] in (2,3,6,7):
                return
            exec(f'self.at_weather_type{identifier[-1]}_wf = block[2:4]')
            exec(f'self.at_weather_ab_type{identifier[-1]}_wf = block[4:8]')

        elif identifier in ('AU1', 'AU2', 'AU3', 'AU4', 'AU5', 'AU6', 'AU7', 'AU8', 'AU9'):
            # discard flag if its qa code is defined as invalid
            if block[-1:] in (2,3,6,7):
                return
            exec(f'self.au_proximity{identifier[-1]}_wf = block[0]')
            exec(f'self.au_descriptor{identifier[-1]}_wf = block[1]')
            exec(f'self.au_precip_code{identifier[-1]}_wf = block[2:4]')
            exec(f'self.au_obscuration{identifier[-1]}_wf = block[4]')
            exec(f'self.au_other_weather{identifier[-1]}_wf = block[5]')
            exec(f'self.au_combination{identifier[-1]}_wf = block[6]')
            if print_info:
                print(isd_map[identifier][map_value])

        elif identifier in ('AW1', 'AW2', 'AW3', 'AW4'):
            if print_info:
                print(isd_map[identifier][map_value])
            # discard flag if its qa code is defined as invalid
            if block[-1:] in (2,3,6,7):
                return
            exec(f'self.aw_auto_atm_condition{identifier[-1]}_wf = block[0:2]')

        elif identifier in ('MV1', 'MV2', 'MV3', 'MV4', 'MV5', 'MV6', 'MV7'):
            if print_info:
                print(isd_map[identifier][map_value])
            # discard flag if its qa code is defined as invalid
            if block[-1:] in (6,7):
                return
            exec(f'self.mv_atm_condition{identifier[-1]}_wf = block[0:2]')

        elif identifier in ('MW1', 'MW2', 'MW3', 'MW4', 'MW5', 'MW6', 'MW7'):
            if print_info:
                print(isd_map[identifier][map_value])
            # discard flag if its qa code is defined as invalid
            if block[-1:] in (2,3,6,7):
                return
            exec(f'self.mw_man_atm_condition{identifier[-1]}_wf = block[0:2]')

        if print_block:
            print('read following block: {}'.format(block))

    def parse_rem_section(self, block):
        """Function to read the Remarks Data Section"""

        # initialize position, after REM and fist id
        info = self.section_chars * 2
        # read how many chars the REM has
        to_read = int(block[info:info + self.section_chars]) + self.section_chars * 3  # the ids we read already
        # start position
        pos = 0
        self.rem_content = block[pos:to_read]

        # # sanity check
        # print('block: {}'.format(block))
        # print('rem content: {}'.format(rem_content))
        return self.rem_content

    def parse_eqd_section(self, block):
        """Function to read the Element Quality Data Section"""

        param_code = 6  # length of the original value text & parameters' code
        rcode_length = 1  # reason code length
        pos = self.section_chars  # starting from 3, after reading section id
        read = self.section_chars

        while pos < len(block):
            # quality identifier
            qid = block[pos:pos + self.section_chars]
            # if qid == 'QNN':
            #     raise ValueError
            # update position
            pos += self.section_chars
            # original value text
            orig_text = block[pos:pos + param_code]
            # update position
            pos += param_code
            # reason code
            reason_code = block[pos:pos + rcode_length]
            # update position
            pos += rcode_length
            # parameter identifier
            parameter = block[pos:pos + param_code]
            # update position
            pos += param_code

            # characters read so far
            read += len(qid + orig_text + reason_code + parameter)

            if 'WOSPD' in parameter:
                self.speed_orig_value = orig_text
                self.speed_reason_code = reason_code

            if 'WODIR' in parameter:
                self.direction_orig_value = orig_text
                self.direction_reason_code = reason_code

            if 'ATOD' in parameter:
                self.dpoint_orig_value = orig_text
                self.dpoint_reason_code = reason_code

            if 'ATOT' in parameter:
                self.temp_orig_value = orig_text
                self.temp_reason_code = reason_code

            if 'APOSLP' in parameter:
                self.pressure_orig_value = orig_text
                self.pressure_reason_code = reason_code

            if 'SCOCIG' in parameter:
                self.ceiling_orig_value = orig_text
                self.ceiling_reason_code = reason_code

            if 'VODIS' in parameter:
                self.distance_orig_value = orig_text
                self.distance_reason_code = reason_code

        return read

    def parse_qnn_section(self):
        self.has_qnn = True
        # logging.info('You will not believe this! This line has a QNN section! Filename: '.format(self.filename[-17:0]))

